/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Save File to.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


var EXPORTED_SYMBOLS =
[
    "PreferenceProxy",
    "PreferencesProxy"
];

const Ci = Components.interfaces;
const Cu = Components.utils;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

Cu.import("resource://savefileto/common.jsm");


function prefGetter(prefs, name)
{
    switch (prefs.getPrefType(name))
    {
        case Ci.nsIPrefBranch.PREF_BOOL:
            return function() { return prefs.getBoolPref(name); };
        case Ci.nsIPrefBranch.PREF_INT:
            return function() { return prefs.getIntPref(name); };
        case Ci.nsIPrefBranch.PREF_STRING:
            return function() { return prefs.getComplexValue(name, Ci.nsISupportsString).data; };
        default:
            throw new Error("unknown pref type: " + prefs.getPrefType(name) + " (" + prefs.root + name + ")");
    }
}


function cachedGetter(getter, name)
{
    return function()
    {
        if (! (name in this))
            this[name] = getter();
        return this[name];
    }
}


function PreferencesProxy(branch, pref2prop)
// extends ListenerManager
// implements nsIObserver
{
    ListenerManager.call(this);
    
    var prefs = Services.prefs.getBranch(branch);
    prefs.QueryInterface(Ci.nsIPrefBranch2);
    prefs.addObserver("", this, true);  // weak reference
    
    this._pref2prop = {};
    for each (let [prefName, propName] in Iterator(pref2prop))
    {
        var hasPropName = (typeof(propName) == "string");
        propName = (hasPropName ? propName : prefName.replace(/\.\w/g, function(s) s[1].toUpperCase()));
        
        this._pref2prop[prefName] = propName;
        this.__defineGetter__(propName, cachedGetter(prefGetter(prefs, prefName), "_" + propName));
    }
}

PreferencesProxy.prototype =
{
    QueryInterface: XPCOMUtils.generateQI([
        Ci.nsIObserver,
        Ci.nsISupports,
        Ci.nsISupportsWeakReference]),
    
    // nsIObserver
    observe: function(subject, topic, data)
    {
        var propName = this._pref2prop[data];
        if (propName)
        {
            delete this["_" + propName];
            this.notifyListeners(this, propName);
        }
    }
}

extend(PreferencesProxy, ListenerManager);


function PreferenceProxy(prefName)
// extends PreferencesProxy
{
    PreferencesProxy.call(this, prefName, {"": "value"});
}

extend(PreferenceProxy, PreferencesProxy);
